#!/usr/bin/python

""" The whole point of input specifications is that they'll contain more
information than is available elsewhere. This begs the question of what kind of
process should be used to annotate the existing resources, in this case the text
file containing XPath locators for each significant node in each of the UBL
document types.

The XPath data files look something like this:

1   1..1 /rs:OrderResponseSimple/
2   1..1 /rs:OrderResponseSimple/rs:ID
2.1 0..1 /rs:OrderResponseSimple/rs:ID/@identificationSchemeAgencyID

Since UBL uses such consistent names for elements and attributes, it's possible
to compactly encode a set of rules that applies to the 160-some unique elements
and attributes used by the system. The rules are defined in IS_Rules.py

Processing combines these data into a single specification. It separately
computes reusable (cac:) groups, as well as those unique to a specific UBL
document type.

For now, the order in the XPath file is taken as the recommended order for
navigation. This might even be correct.

Next steps:

TODO: intelligently deal with cardinality, so that 1..7 for example gets treated
as a repeating section

TODO: In the recursion, everything that nests is currently split out. Need an
exception list for things to flatten

TODO: Identify places that actually use lists, then list possible values

TODO: Figure out what code lists make sense to represent here.


"""

import sys
import IS_Rules

class blank: pass
ubldata = blank()
ubldata.doctype = {}
ubldata.annot = {}
ubldata.all_leaves = {}
ubldata.marked = {}
ubldata.flatten = {}

path = "xpaths/text/"

files = [ "DespatchAdvice", "Invoice", "Order", "OrderCancellation",
          "OrderChange", "OrderResponse", "OrderResponseSimple", "ReceiptAdvice" ]

#files = [ "OrderResponseSimple" ]

suffix = "-XPath.txt"

def main():
    gather()
    report()

def gather():
    "read in all needed data into handy data structures"
    for file in files:
        entries = []
        f = open(path + file + suffix)
        for line in f:
            # parse and get into a self-documenting data structure
            if line[0] != "#":
                tokens = line.split()
                record = blank()
                record.numref = tokens[0]
                record.cardinality = tokens[1]
                record.card_min = int(tokens[1].split("..")[0])
                try:
                    record.card_max = int(tokens[1].split("..")[1])
                except ValueError:
                    record.card_max = sys.maxint
                record.path = tokens[2]
                record.pathsplit = record.path.split("/")
                record.leaf = record.pathsplit[-1]
                record.terminal = True
                if record.leaf == "":
                    record.leaf = record.pathsplit[-2] + "/"
                    record.terminal = False
                if ubldata.all_leaves.has_key(record.leaf):
                    ubldata.all_leaves[record.leaf] = ubldata.all_leaves[record.leaf] + 1
                else:
                    ubldata.all_leaves[record.leaf] = 1
                record.root = record.pathsplit[1]
                record.root2 = record.pathsplit[2]
                entries.append(record)
        ubldata.doctype[file] = entries
        f.close()

    # ------- rules -------
    al2 = {}
    for leaf in ubldata.all_leaves:
        al2[leaf] = ""

    def rule(match, value):
        for leaf in al2:
            if leaf.find(match) > -1:
                #if al2[leaf] != "" and al2[leaf] != value:
                #    print "changing " + leaf + " from " + al2[leaf] + " to " + value
                al2[leaf] = value

    IS_Rules.configure_rules(rule)

    # integrity check
    for leaf in al2:
        if al2[leaf]=="":
            #if you get this message, stop and figure out what rules are missing
            print "!!!!!  Emptiness at " + leaf

    ubldata.annot = al2

    # ------- flatten groups ------
    fg = {}
    def flatten(name):
        fg[name] = ""

    IS_Rules.configure_flattened_groups(flatten)

    ubldata.flatten = fg

def report():
    "write out the processed data"
    # file-specific reports
    for file in files:
        print
        print file
        print "============================"
        print "Summary of " + file
        grps = {}
        rep_group(file, ubldata.doctype[file][0].root , 2, grps)
        # handle cac: groups elsewhere
        grps = [item for item in grps if item[2]==":"]
        if grps:
            print
            print "Local groups for " + file
            print "-----------------------------"
            rep_localgroups(file, grps, 3)

    # any overall reporting goes here
    print "===== Global  Groups ======="
    subgrps = {}
    for file in files: # first get a list of all the 2nd-level groups
        for record in ubldata.doctype[file]:
            if not record.terminal and len(record.root2)> 3 and record.root2[3]==":":
                subgrps[record.root2] = file

    level = 3
    while len(subgrps) > 0:
        print str(level) + "---"
        s = subgrps
        subgrps = {}
        for grp in s:
            if not ubldata.marked.has_key(grp):
                print grp + ":"
                rep_group(s[grp], grp, level, subgrps)
                print
        level = level + 1



def rep_localgroups(file, grps, level):
    for grp in grps:
        if not ubldata.marked.has_key(grp):
            print grp + ":"
            rep_group(file, grp, level, {})
            print

def rep_group(file, grp, level, subgrps, indent=""):
    ubldata.marked[grp] = file
    lastrootN = ""
    inPlay = 0  # 0->Not started yet, 1->Working, 2->Finished
    #avoids situation where a group is used more than once within a file
    for record in ubldata.doctype[file]:
        if grp in record.pathsplit and inPlay != 2:
            if inPlay==0: inPlay = 1
            try:
                rootN = record.pathsplit[level]
            except IndexError:
                continue
            if record.leaf == grp + "/": # dont match blah/blah/grp/
                continue                 # opposed to blah/blag/grp/child...
            if rootN != lastrootN:
                lastrootN = rootN
                reqstr = ""
                if record.card_min > 0: reqstr = " (required)"
                dt = ubldata.annot[record.leaf]
                if dt != "hidden":
                    print indent + record.cardinality + " " + record.leaf + " " + dt + reqstr

                if dt=="group" and ubldata.flatten.has_key(record.leaf[:-1]):
                    print indent + "  --- subgroup ---:"
                    rep_group(file, record.leaf[:-1], level+1, subgrps, indent+"  ")


                if not record.terminal:
                    subgrps[record.pathsplit[level]] = file
        else:
            if inPlay==1: inPlay = 2


if __name__ == '__main__':
    main()

#end